@@ -12,27 +12,34 @@ import darkdetect
 import threading
 import traceback
 import platform
-import requests
+import copy
 import json
 import sys
 import io
 import re
 import os
 
-from PIL import Image, ImageTk
+from PIL import Image
 
-from typing import List, Tuple, Any
+from typing import List, Tuple, Any, Optional
+
+import nb2pb
 
 import netsblox
 from netsblox import transform
 from netsblox import common
+from netsblox import rooms
 
 NETSBLOX_PY_PATH = os.path.dirname(netsblox.__file__)
 
 SUGGESTION_UPDATE_INTERVAL = 200
+PANED_WINDOW_OPTS = {
+    'sashwidth': 5,
+}
 
 xux = lambda x: f'{x} {x.upper()}'
-PROJECT_FILETYPES = [('Project Files', xux('.json')), ('All Files', '.*')]
+PROJECT_FILETYPES = [('PyBlox Project Files', xux('.json')), ('All Files', '.*')]
+NB_PROJECT_FILETYPES = [('NetsBlox Project Files', xux('.xml')), ('All Files', '.*')]
 PYTHON_FILETYPES = [('Python Files', xux('.py')), ('All Files', '.*')]
 IMAGE_FILETYPES = [('Images', xux('.png .jpg .jpeg')), ('All Files', '.*')]
 
@@ -79,6 +86,7 @@ else:
         'right-click': 'Button-3',
     }
 
+nb = None
 root = None
 main_menu = None
 content = None
@@ -297,15 +305,14 @@ class Content(tk.Frame):
         super().__init__(parent)
         self.pack(side = tk.BOTTOM, fill = tk.BOTH, expand = True)
 
-        self.project = ProjectEditor(self)
-        self.display = Display(self)
+        self.pane = tk.PanedWindow(self, orient = tk.HORIZONTAL, **PANED_WINDOW_OPTS)
+        self.pane.pack(fill = tk.BOTH, expand = True)
 
-        self.project.grid(row = 0, column = 0, sticky = tk.NSEW)
-        self.display.grid(row = 0, column = 1, sticky = tk.NSEW)
+        self.project = ProjectEditor(self.pane)
+        self.display = Display(self.pane)
 
-        self.grid_columnconfigure(0, weight = 5, uniform = 'content')
-        self.grid_columnconfigure(1, weight = 3, uniform = 'content')
-        self.grid_rowconfigure(0, weight = 1)
+        self.pane.add(self.project, stretch = 'always', width = 5, minsize = 300)
+        self.pane.add(self.display, stretch = 'always', width = 3, minsize = 300)
 
 class DndTarget:
     def __init__(self, widget, on_start, on_stop, on_drop):
@@ -398,73 +405,8 @@ class BlocksList(tk.Frame):
         self.text.configure(state = tk.DISABLED)
 
 class Imports:
-    RAW_INFO = [
-        ['time', 'time', '''
-            The `time` package brings in several utilities related to measuring time.
+    RAW_INFO = json.loads(common.load_text("netsblox://assets/default-imports.json"))
 
-            Here is a list of especially useful functions:
-             - `time.sleep(t)` lets your scripts wait a certain number of seconds before resuming.
-             - `time.time()` gives you the current timestamp. You can subtract two times to get the time interval in seconds.
-        '''],
-        ['math', 'math', '''
-            The `math` module brings in many common mathematical functions such as `sin`, `cos`, `tan`,
-            inverse trig, exponentiation, and many more specialized tools.
-
-            This is useful if you need to compute a handful of operations in your scripts.
-            However, if you need to do a lot of number crunching, consider using the `numpy` package instead.
-        '''],
-        ['random', 'random', '''
-            The `random` module allows you to generate random numbers.
-
-            Here is a list of especially useful functions:
-             - `random.random()` gives you a float (decimal) number between 0.0 and 1.0 (not including 1.0).
-             - `random.randrange(a,b)` gives you a random integer in `range(a,b)`.
-            Just like the `range` function, you can pass a single argument to default the first to 0.
-        '''],
-        ['fractions', 'frac', '''
-            The `fractions` module (`frac` for short) allows you to do math directly on rational numbers (fractions).
-
-            If you use built-in types like `float` to do numeric calculations, you might encounter some amount of rounding error.
-            This is because `float` has a maximum number of "significant digits" it can represent, meaning it cannot store all numbers exactly.
-            For instance, the number `1/3` written in binary has an infinite repeating patterns of 1's and 0's, so `float` cannot store it exactly.
-            On the other hand, `frac.Fraction` stores the number as a pair of integers (numerator and denominator), so fractions can be stored exactly.
-
-            To create a fraction, you can use `frac.Fraction(my_number)` to perform the conversion.
-            Fractions support the typical math operators, so you can write (frac.Fraction(773) / frac.Fraction(3))**2 and it will give you
-            back a fraction containing exactly 597529/9 (the same calculation using `float` has an error of `1.4552e-11`).
-            Although rounding error from individual operations on `float` are small, they add up when performing many calculations.
-            `frac.Fraction` can be used to avoid any error, so long as you only use operations that work with rational numbers
-            (e.g., you cannot take the cos of a fraction, but you can add, subtract, multiply, divide, raise to integer powers, etc.).
-
-            Printing out a fraction will show the value in fraction form.
-            If you want to convert the fraction into a `float` (e.g., to display it as a decimal),
-            you can use `float(my_frac)` to perform the conversion.
-        '''],
-        ['numpy', 'np', '''
-            The `numpy` package is a highly specialized tool for doing a lot of number crunching very very quickly.
-            By convention, `numpy` is typically imported as `np` for brevity.
-
-            Most `numpy` functions work on arrays (lists). To make a `numpy` array, simply use `np.array(some_list)`.
-            `numpy` arrays perform their operations in batches.
-            For instance, if you have two `numpy` arrays `p` and `q`, each of which is an array of three numbers like 3D coordinates,
-            you can simply write `p+q` and numpy will add the components together and give you a new array.
-            Many operations are supported, for instance `np.sin(p + 2 * q**2)` will square all the elements of `q`, double them,
-            add them to the elements of `p`, then take the `sin` of each element and give the result as a new array.
-
-            `numpy` arrays are actually multi-dimensional arrays.
-            A 1D array is a "normal" array (list).
-            A 2D array is a matrix.
-            In higher dimensions (and in general) these are called tensors, and are commonly used in advanced applications like machine learning.
-            If you pass `np.array()` a list of lists, you will get an N-dimensional array (tensor) with the same structure.
-
-            Python supports list slicing. For instance `p[3:5]` gets the sublist of elements with index 3 up to (but not including) 5.
-            However, python lists do not support higher-dimensional slicing (even if they are lists of lists).
-            But `numpy` arrays allow this, with each dimension separated by a comma.
-            For instance if `p` is a 5x5 matrix, `p[1,2:4]` gives row index 1 elements 2 up to 4, and `p[1:-1,1:-1]` gives you the 3x3 matrix in the middle.
-
-            As you can see, `numpy` lets you do a lot of work very quickly and without much typing.
-        '''],
-    ]
     def __init__(self, *, on_update = None):
         self.packages = {}
         for item in Imports.RAW_INFO:
@@ -517,6 +459,11 @@ class ProjectEditor(tk.Frame):
         # it is not stored in project files or exports because conflicting ids would break messaging.
         self.project_id = common.generate_proj_id()
 
+        self.roles: List[dict] = None
+        self.active_role: int = None
+
+        self.block_sources = []
+
         self.editors: List[CodeEditor] = []
         self.show_blocks_tkvar = tk.BooleanVar()
 
@@ -658,34 +605,68 @@ class ProjectEditor(tk.Frame):
 
     def get_save_dict(self) -> dict:
         res = {}
-        res['blocks'] = {
-            'sources': self.block_sources[:],
-            'global': [x.copy() for x in GlobalEditor.blocks if x['source'] == None],
-            'stage':  [x.copy() for x in  StageEditor.blocks if x['source'] == None],
-            'turtle': [x.copy() for x in TurtleEditor.blocks if x['source'] == None],
-        }
+
+        res['roles'] = []
+        for i, role in enumerate(self.roles):
+            if i != self.active_role:
+                res['roles'].append(copy.deepcopy(role))
+                continue
+
+            role_res = {}
+            res['roles'].append(role_res)
+
+            role_res['name'] = role['name']
+            role_res['block_sources'] = self.block_sources[:]
+            role_res['blocks'] = {
+                'global': [x.copy() for x in GlobalEditor.blocks if x['source'] == None],
+                'stage':  [x.copy() for x in  StageEditor.blocks if x['source'] == None],
+                'turtle': [x.copy() for x in TurtleEditor.blocks if x['source'] == None],
+            }
+
+            role_res['imports'] = []
+            for pkg, item in self.imports.packages.items():
+                if item['tkvar'].get():
+                    role_res['imports'].append(pkg)
+
+            role_res['editors'] = []
+            for editor in self.editors:
+                ty = None
+                if isinstance(editor, GlobalEditor): ty = 'global'
+                elif isinstance(editor, StageEditor): ty = 'stage'
+                elif isinstance(editor, TurtleEditor): ty = 'turtle'
+                else: raise Exception(f'unknown editor type: {type(editor)}')
+                role_res['editors'].append({
+                    'type': ty,
+                    'name': editor.name,
+                    'value': editor.text.get('1.0', 'end-1c'),
+                })
+
+            role_res['images'] = { name: common.encode_image(img) for name, img in self.imports.images.items() }
+
         res['show_blocks'] = self.show_blocks
-        res['imports'] = []
-        for pkg, item in self.imports.packages.items():
-            if item['tkvar'].get():
-                res['imports'].append(pkg)
-        res['editors'] = []
-        for editor in self.editors:
-            ty = None
-            if isinstance(editor, GlobalEditor): ty = 'global'
-            elif isinstance(editor, StageEditor): ty = 'stage'
-            elif isinstance(editor, TurtleEditor): ty = 'turtle'
-            else: raise Exception(f'unknown editor type: {type(editor)}')
-            res['editors'].append({
-                'type': ty,
-                'name': editor.name,
-                'value': editor.text.get('1.0', 'end-1c'),
-            })
-        res['images'] = { name: common.encode_image(img) for name, img in self.imports.images.items() }
+
         return res
-    def load(self, proj: dict) -> None:
+    def load(self, *, super_proj: Optional[dict] = None, active_role: Optional[int] = None) -> None:
+        if super_proj is None:
+            roles = self.roles
+        else:
+            super_proj = copy.deepcopy(super_proj)
+            roles = super_proj.get('roles', [super_proj]) # default is for backward compat
+
+        if active_role is None:
+            active_role = 0
+        elif active_role < 0 or active_role >= len(roles):
+            active_role = 0
+        proj = roles[active_role]
+
+        counter = 0
+        for role in roles:
+            if 'name' in role: continue
+            role['name'] = 'untitled' if counter == 0 else f'untitled_{counter}'
+            counter += 1
+
         new_blocks = { 'global': [], 'stage': [], 'turtle': [] }
-        new_sources = proj.get('blocks', {}).get('sources', [])
+        new_sources = proj.get('block_sources', [])
 
         def add_blocks(blocks, *, source):
             for k, v in blocks.items():
@@ -697,13 +678,16 @@ class ProjectEditor(tk.Frame):
 
         for src in new_sources:
             add_blocks(json.loads(common.load_text(src)), source = src)
-        add_blocks({ k: v for k, v in proj.get('blocks', {}).items() if k in new_blocks }, source = None)
+        add_blocks(proj.get('blocks', {}), source = None)
         add_blocks({ # legacy support
             'global': proj.get('global_blocks', []),
             'stage': proj.get('stage_blocks', []),
             'turtle': proj.get('turtle_blocks', []),
         }, source = None)
 
+        self.roles = roles
+        self.active_role = active_role
+
         GlobalEditor.blocks = new_blocks['global']
         StageEditor.blocks = new_blocks['stage']
         TurtleEditor.blocks = new_blocks['turtle']
@@ -714,7 +698,8 @@ class ProjectEditor(tk.Frame):
             self.editors[i].destroy()
         self.editors = []
 
-        self.show_blocks = proj['show_blocks']
+        if super_proj is not None:
+            self.show_blocks = super_proj['show_blocks']
 
         for info in proj['editors']:
             ty = info['type']
@@ -734,8 +719,11 @@ class ProjectEditor(tk.Frame):
             item['tkvar'].set(False)
         for pkg in proj['imports']:
             self.imports.packages[pkg]['tkvar'].set(True)
+
+        self.imports.images.clear()
         for name, raw in proj.get('images', {}).items():
             self.imports.images[name] = common.decode_image(raw)
+
         self.imports.batch_update()
 
 class ContextMenu(tk.Menu):
@@ -810,8 +798,11 @@ class ChangedText(tk.Text):
         try:
             result = self.tk.call(cmd)
         except Exception as e:
-            # for some reason our proxying breaks undo/redo if the stacks are empty, so just catch and ignore
-            if cmd[1:] not in [('edit', 'undo'), ('edit', 'redo')]:
+            # for some reason our proxying breaks some ops in some cases, so just catch and ignore
+            ignore_patterns = [
+                ('edit', 'undo'), ('edit', 'redo'), # fails if stack is empty (undo) or at top of stack (redo)
+            ]
+            if cmd[1:] not in ignore_patterns:
                 raise e
 
         # generate an event if something was added or deleted, or the cursor position changed
@@ -870,14 +861,6 @@ class ScrolledText(tk.Frame):
         if readonly:
             # make text readonly be ignoring all (default) keystrokes
             self.text.bind('<Key>', lambda e: 'break')
-
-            # catching all keys above means we can't copy anymore - impl manually
-            def on_copy(e):
-                self.clipboard_clear()
-                self.clipboard_append(self.text.selection_get())
-                return 'break'
-            self.text.bind(f'<{SYS_INFO["mod"]}-Key-c>', on_copy)
-            self.text.bind(f'<{SYS_INFO["mod"]}-Key-C>', on_copy)
         else:
             def on_redo(e):
                 self.text.edit_redo()
@@ -896,6 +879,17 @@ class ScrolledText(tk.Frame):
             self.text.bind(f'<{SYS_INFO["mod"]}-Key-v>', on_paste)
             self.text.bind(f'<{SYS_INFO["mod"]}-Key-V>', on_paste)
 
+        # custom copy behavior - (also in readonly case above, catching all keys means we can't copy without override anyway)
+        def on_copy(e):
+            self.clipboard_clear()
+            if self.text.tag_ranges(tk.SEL):
+                self.clipboard_append(self.text.get(tk.SEL_FIRST, tk.SEL_LAST))
+            else: # if no selection, copy the current line
+                self.clipboard_append(self.text.get('insert linestart', 'insert lineend'))
+            return 'break'
+        self.text.bind(f'<{SYS_INFO["mod"]}-Key-c>', on_copy)
+        self.text.bind(f'<{SYS_INFO["mod"]}-Key-C>', on_copy)
+
         if linenumbers:
             self.linenumbers = TextLineNumbers(self, target = self.text, width = 40)
             self.text.bind('<<Change>>', lambda e: self.on_content_change())
@@ -1101,7 +1095,7 @@ class CodeEditor(ScrolledText):
         if should_show:
             if self.help_popup is not None:
                 self.help_popup.destroy()
-            
+
             try:
                 x, y, w, h = self.text.bbox(tk.INSERT)
             except:
@@ -1117,7 +1111,7 @@ class CodeEditor(ScrolledText):
                 if not item.name.startswith('_'): # hide private stuff - would only confuse beginners and they shouldn't touch it anyway
                     self.help_popup.insert(tk.END, item.name)
                     self.help_completions[item.name] = item.complete
-            
+
             self.help_popup.bind('<Double-Button-1>', lambda e: self.do_completion())
         else:
             self.hide_suggestion()
@@ -1156,7 +1150,7 @@ class CodeEditor(ScrolledText):
         self.text.delete(*sel_padded)
         self.text.insert(sel_padded[0], mutated)
         self.text.edit_separator()
-        
+
         new_sel_start = f'{sel_start_pieces[0]}.{max(0, int(sel_start_pieces[1]) + line_deltas[0])}'
         new_sel_end = f'{sel_end_pieces[0]}.{max(0, int(sel_end_pieces[1]) + line_deltas[-1])}'
         new_ins = f'{ins_pieces[0]}.{max(0, int(ins_pieces[1]) + ins_delta)}'
@@ -1212,7 +1206,8 @@ from netsblox.turtle import *
 from netsblox.concurrency import *
 nb = netsblox.Client(proj_name = """$proj_name""", proj_id = $proj_id)
 'A connection to NetsBlox, which allows you to use services and RPCs from python.'
-getattr(netsblox.turtle._get_proj_handle(), '_Project__tk').title(f'PyBlox - {nb.get_public_id()}')
+getattr(netsblox.turtle._get_proj_handle(), '_Project__tk').title(f'PyBlox - {nb.public_id}')
+nb.set_room($room_handle)
 setup_stdio()
 setup_yielding()
 import time as _time
@@ -1221,7 +1216,7 @@ def _yield_(x):
     return x
 
 '''.lstrip()
-    BASE_PREFIX_LINES = 13
+    BASE_PREFIX_LINES = 14
 
     prefix = BASE_PREFIX
     prefix_lines = BASE_PREFIX_LINES
@@ -1236,6 +1231,13 @@ def _yield_(x):
         pre = GlobalEditor.prefix
         pre = pre.replace('$proj_name', main_menu.project_name)
         pre = pre.replace('$proj_id', 'None' if is_export else f'\'{content.project.project_id}\'')
+
+        room = main_menu.room_manager
+        role_str = f'\'{content.project.roles[content.project.active_role]["name"]}\''
+        room_id_str = f'\'{room.room_id}\''
+        password_str = 'None' if room.room_password is None else f'\'{room.room_password}\''
+        pre = pre.replace('$room_handle', 'None' if is_export or room.room_name is None else f'netsblox.rooms.RuntimeRoomManager(client = nb, role = {role_str}, room_id = {room_id_str}, password = {password_str})')
+
         return pre + self.text.get('1.0', 'end-1c')
 
 class StageEditor(CodeEditor):
@@ -1268,17 +1270,16 @@ class Display(tk.Frame):
     def __init__(self, parent):
         super().__init__(parent)
 
-        self.docs = ScrolledText(self, readonly = True)
-        self.terminal = TerminalOutput(self)
+        self.pane = tk.PanedWindow(self, orient = tk.VERTICAL, **PANED_WINDOW_OPTS)
+        self.pane.pack(fill = tk.BOTH, expand = True)
+
+        self.docs = ScrolledText(self.pane, readonly = True)
+        self.terminal = TerminalOutput(self.pane)
 
         self.docs.text.configure(wrap = tk.WORD)
 
-        self.docs.grid(row = 0, column = 0, sticky = tk.NSEW)
-        self.terminal.grid(row = 1, column = 0, sticky = tk.NSEW)
-
-        self.grid_columnconfigure(0, weight = 1)
-        self.grid_rowconfigure(0, weight = 1, uniform = 'display')
-        self.grid_rowconfigure(1, weight = 1, uniform = 'display')
+        self.pane.add(self.docs, stretch = 'always', minsize = 100)
+        self.pane.add(self.terminal, stretch = 'always', minsize = 100)
 
 class TerminalOutput(tk.Frame):
     def __init__(self, parent):
@@ -1306,7 +1307,7 @@ class TerminalOutput(tk.Frame):
                     self.old.write(data)
                     self.old.flush()
                 _print_queue.put(data)
-            
+
             def flush(self):
                 pass
             def __len__(self):
@@ -1355,12 +1356,26 @@ class MainMenu(tk.Menu):
 
         def kill():
             if self.try_close_project():
+                self.room_manager.destroy()
                 root.destroy()
         root.protocol('WM_DELETE_WINDOW', kill)
 
         submenu = tk.Menu(self, **MENU_STYLE)
-        submenu.add_command(label = 'New', command = self.new_project, accelerator = f'{SYS_INFO["mod-str"]}+N')
+        submenu.add_command(label = 'New', command = lambda: self.open_project(super_proj = ProjectEditor.DEFAULT_PROJECT), accelerator = f'{SYS_INFO["mod-str"]}+N')
         submenu.add_command(label = 'Open', command = self.open_project, accelerator = f'{SYS_INFO["mod-str"]}+O')
+        submenu.add_command(label = 'Import NetsBlox project', command = self.open_trans_xml)
+
+        subsubmenu = tk.Menu(submenu, **MENU_STYLE)
+        for file in sorted(os.listdir(f'{common._NETSBLOX_PY_PATH}/assets/examples/')):
+            def make_opener(file):
+                def opener():
+                    with open(f'{common._NETSBLOX_PY_PATH}/assets/examples/{file}') as f:
+                        content = json.load(f) # don't cache projects (could be large)
+                    self.open_project(super_proj = content)
+                return opener
+            subsubmenu.add_command(label = file[:-5], command = make_opener(file))
+
+        submenu.add_cascade(label = 'Example', menu = subsubmenu)
         submenu.add_separator()
         submenu.add_command(label = 'Save', command = self.save, accelerator = f'{SYS_INFO["mod-str"]}+S')
         submenu.add_command(label = 'Save As', command = self.save_as, accelerator = f'Shift+{SYS_INFO["mod-str"]}+S')
@@ -1370,7 +1385,7 @@ class MainMenu(tk.Menu):
         submenu.add_command(label = 'Exit', command = kill)
         self.add_cascade(label = 'File', menu = submenu)
 
-        root.bind_all(f'<{SYS_INFO["mod"]}-n>', lambda e: self.new_project())
+        root.bind_all(f'<{SYS_INFO["mod"]}-n>', lambda e: self.open_project(super_proj = ProjectEditor.DEFAULT_PROJECT))
         root.bind_all(f'<{SYS_INFO["mod"]}-o>', lambda e: self.open_project())
         root.bind_all(f'<{SYS_INFO["mod"]}-s>', lambda e: self.save())
         root.bind_all(f'<{SYS_INFO["mod"]}-S>', lambda e: self.save_as())
@@ -1381,6 +1396,11 @@ class MainMenu(tk.Menu):
 
         root.bind_all(f'<{SYS_INFO["mod"]}-b>', lambda e: self.toggle_blocks())
 
+        self.room_manager = rooms.EditorRoomManager(client = nb)
+
+        self.roles_dropdown = tk.Menu(self, **MENU_STYLE)
+        self.add_cascade(label = 'Roles', menu = self.roles_dropdown)
+
         submenu = tk.Menu(self, **MENU_STYLE)
         imp = content.project.imports
         imp_packages = list(imp.packages.items())
@@ -1448,6 +1468,7 @@ class MainMenu(tk.Menu):
                 with open(self.project_path, 'w') as f:
                     json.dump(save_dict, f, separators = (', ', ': '), indent = 2)
                 self.saved_project_dict = save_dict
+                content.project.roles = save_dict['roles'] # sync the in-memory role content
                 return True
             except Exception as e:
                 messagebox.showerror('Failed to save project', str(e))
@@ -1471,38 +1492,58 @@ class MainMenu(tk.Menu):
             except Exception as e:
                 messagebox.showerror('Failed to save exported project', str(e))
 
-    def new_project(self):
-        if not self.try_close_project():
-            return
-
-        content.project.load(ProjectEditor.DEFAULT_PROJECT)
-        self.project_path = None
-        self.saved_project_dict = None
-
-    def open_project(self):
+    def open_project(self, *, super_proj: Optional[dict] = None, active_role: Optional[int] = None):
         content.project.on_tab_change()
-
-        if not self.try_close_project():
-            return
-
-        p = filedialog.askopenfilename(filetypes = PROJECT_FILETYPES)
-        if type(p) is not str or not p:
-            return
+        if not self.try_close_project(): return
 
         rstor = None
+        p = None
         try:
-            with open(p, 'r') as f:
-                proj = json.load(f)
+            if super_proj is None:
+                p = filedialog.askopenfilename(filetypes = PROJECT_FILETYPES)
+                if type(p) is not str or not p:
+                    return
+                with open(p, 'r') as f:
+                    super_proj = json.load(f)
+
+            if content.project.roles is not None:
                 rstor = content.project.get_save_dict() # in case load fails
-                content.project.load(proj)
-                self.project_path = p
-                self.saved_project_dict = proj
+
+            content.project.load(super_proj = super_proj, active_role = active_role)
+            self.saved_project_dict = super_proj
+            self.project_path = p
+            self.update_roles()
         except Exception as e:
             messagebox.showerror('Failed to load project', str(e))
             if rstor is not None:
                 content.project.load(rstor)
 
+    def open_trans_xml(self):
+        content.project.on_tab_change()
+        if not self.try_close_project(): return
+
+        try:
+            p = filedialog.askopenfilename(filetypes = NB_PROJECT_FILETYPES)
+            if type(p) is not str or not p:
+                return
+            with open(p, 'r') as f:
+                xml = f.read()
+            proj = json.loads(nb2pb.translate(xml)[1])
+            self.open_project(super_proj = proj)
+        except Exception as e:
+            messagebox.showerror('Failed to import project', str(e))
+
+    def switch_role(self, *, active_role: int):
+        content.project.on_tab_change()
+        if not self.try_close_project(): return
+
+        content.project.load(active_role = active_role)
+        self.update_roles()
+
     def try_close_project(self) -> bool: # true if user accepted close
+        if self.saved_project_dict is None:
+            return True # fires (only) on first initialization
+
         save_dict = content.project.get_save_dict()
         if save_dict == self.saved_project_dict:
             return True # if saved project content is equal, no need to do anything
@@ -1530,14 +1571,13 @@ class MainMenu(tk.Menu):
 
         name = None
         while True:
-            name = simpledialog.askstring('Name Image', 'Enter the name of the image, which is used to access it from code')
-            if name is None:
-                return
+            name = simpledialog.askstring(title = 'Name Image', prompt = 'Enter the name of the image, which is used to access it from code')
+            if name is None: return
             if not is_valid_ident(name):
-                messagebox.showerror(title = 'Invalid name', message = f'"{name}" is not a valid python variable name')
+                messagebox.showerror('Invalid name', message = f'"{name}" is not a valid python variable name')
                 continue
             if name in content.project.imports.images:
-                messagebox.showerror(title = 'Invalid name', message = f'An image named "{name}" already exists')
+                messagebox.showerror(title = 'Invalid name', message = f'An image named {name} already exists')
                 continue
             break
 
@@ -1560,8 +1600,8 @@ class MainMenu(tk.Menu):
 
             def get_deleter(name):
                 def deleter():
-                    title = f'Delete {name}'
-                    msg = f'Are you sure you would like to delete {name}? This operation cannot be undone.'
+                    title = f'Delete sprite {name}'
+                    msg = f'Are you sure you would like to delete sprite {name}? This operation cannot be undone.'
                     if messagebox.askyesno(title, msg, icon = 'warning', default = 'no'):
                         del content.project.imports.images[name]
                         content.project.imports.batch_update()
@@ -1570,8 +1610,101 @@ class MainMenu(tk.Menu):
 
             self.images_dropdown.add_cascade(label = f'{name} {img.width}x{img.height}', menu = submenu)
 
+    def update_roles(self):
+        self.roles_dropdown.delete(0, 'end')
+
+        submenu = tk.Menu(**MENU_STYLE)
+
+        def create_room():
+            password = simpledialog.askstring(title = 'New Room Password', prompt = 'Password for the new room (or nothing for no password)')
+            if password is None: return
+            if password == '': password = None
+            try:
+                self.room_manager.create_room(password)
+                messagebox.showinfo('Created Room', message = f'Created Room\nName: {self.room_manager.room_name}\n\nOther clients can join with\nRole > Room > Join Room')
+            except Exception as e:
+                messagebox.showerror('Failed to create room', message = str(e))
+            self.update_roles()
+        submenu.add_command(label = 'Create New Room', command = create_room)
+
+        def join_room():
+            room_name = simpledialog.askstring(title = 'Room Name', prompt = 'Enter the name of the room to join')
+            if room_name is None or room_name == '': return
+            password = simpledialog.askstring(title = 'Room Password', prompt = f'Password for room \'{room_name}\' (or nothing if no password)')
+            if password is None: return
+            if password == '': password = None
+            try:
+                self.room_manager.join_room(room_name, password)
+                messagebox.showinfo('Joined Room', message = f'Joined Room\nName: {self.room_manager.room_name}')
+            except Exception as e:
+                messagebox.showerror('Failed to join room', message = str(e))
+            self.update_roles()
+        submenu.add_command(label = 'Join Room', command = join_room)
+
+        def leave_room():
+            self.room_manager.leave_room()
+            self.update_roles()
+        submenu.add_command(label = f'Leave Room ({self.room_manager.room_name})' if self.room_manager.room_name is not None else 'Leave Room',
+            command = leave_room, state = tk.ACTIVE if self.room_manager.room_name is not None else tk.DISABLED)
+
+        self.roles_dropdown.add_cascade(label = 'Room', menu = submenu)
+
+        def make_role():
+            name = None
+            while True:
+                name = simpledialog.askstring(title = 'Name Role', prompt = 'Enter the name of the new role, which should be a valid variable name')
+                if name is None: return
+                if not is_valid_ident(name):
+                    messagebox.showerror('Invalid name', message = f'"{name}" is not a valid python variable name')
+                    continue
+                if any(x['name'] == name for x in content.project.roles):
+                    messagebox.showerror(title = 'Invalid name', message = f'A role named {name} already exists.')
+                    continue
+                break
+            role = copy.deepcopy(ProjectEditor.DEFAULT_PROJECT['roles'][0])
+            role['name'] = name
+            content.project.roles.append(role)
+            self.update_roles()
+        self.roles_dropdown.add_command(label = 'New Role', command = make_role)
+
+        if len(content.project.roles) != 0:
+            self.roles_dropdown.add_separator()
+
+        for i, role in enumerate(content.project.roles):
+            submenu = tk.Menu(**MENU_STYLE)
+            is_current = i == content.project.active_role
+
+            def get_switcher(idx):
+                def do_switch():
+                    if idx == content.project.active_role: return # santy check
+                    self.switch_role(active_role = idx)
+                return do_switch
+            submenu.add_command(label = 'Open', command = get_switcher(i), state = tk.DISABLED if is_current else tk.ACTIVE)
+
+            def get_deleter(idx):
+                def do_delete():
+                    if idx == content.project.active_role:
+                        messagebox.showerror('Cannot delete active role', 'To delete this role, open a different role first')
+                        return
+
+                    name = content.project.roles[idx]['name']
+                    title = f'Delete role {name}'
+                    msg = f'Are you sure you would like to delete role {name}? This operation cannot be undone.'
+                    if messagebox.askyesno(title, msg, icon = 'warning', default = 'no'):
+                        del content.project.roles[idx]
+                        if content.project.active_role >= idx:
+                            content.project.active_role -= 1
+                        self.update_roles()
+                return do_delete
+            submenu.add_command(label = 'Delete', command = get_deleter(i), state = tk.DISABLED if is_current else tk.ACTIVE)
+
+            role_name = content.project.roles[i]['name']
+            self.roles_dropdown.add_cascade(label = f'{role_name} (active)' if is_current else role_name, menu = submenu)
+
 def main():
-    global root, main_menu, content
+    global nb, root, main_menu, content
+
+    nb = netsblox.Client()
 
     root = tk.Tk()
     root.geometry('1200x600')
@@ -1587,13 +1720,12 @@ def main():
     main_menu = MainMenu(root)
 
     if len(sys.argv) <= 1:
-        content.project.load(ProjectEditor.DEFAULT_PROJECT)
+        main_menu.open_project(super_proj = ProjectEditor.DEFAULT_PROJECT)
     elif len(sys.argv) == 2:
         with open(sys.argv[1], 'r') as f:
             save_dict = json.load(f)
-            content.project.load(save_dict)
-            main_menu.project_path = os.path.abspath(sys.argv[1])
-            main_menu.saved_project_dict = save_dict
+        main_menu.open_project(super_proj = save_dict)
+        main_menu.project_path = os.path.abspath(sys.argv[1])
     else:
         print(f'usage: {sys.argv[0]} (project)', file = sys.stderr)
         sys.exit(1)
